# max30102.py
# MAX30102-Modul mit MAX30102-Klasse
# Stand: 17.03.2025

# Probleme:
# 1. Mit der Handy-Kamera ist (selbst bei 25,4 mA) kein Leuchten
# bei der IR-LED zu sehen! Defekt? Aber: beim Aktivieren der IR-LED
# "zieht" der MAX-Baustein mehr Strom! VGl. auch 4.
# 2. Wenn RED-LED auf 25,4 mA eingestellt wird, bricht das Programm ab.
# 3. Manchmal tritt die Fehlermeldung "[Errno 19] ENODEV" auf.
# 4. ir_led-Graph von HR1a.py mit I(Red_LED) = 0 und I(IR_LED) = 7 mA 
# ähnelt dem von red_led; d. h. IR-LED leuchtet!!!


from machine import Pin, I2C
from time import sleep

# Register und ihre Adressen 
REG_INTR_STATUS_1 = 0x00
REG_INTR_STATUS_2 = 0x01

REG_INTR_ENABLE_1 = 0x02
REG_INTR_ENABLE_2 = 0x03

REG_FIFO_WR_PTR = 0x04
REG_OVF_COUNTER = 0x05
REG_FIFO_RD_PTR = 0x06
REG_FIFO_DATA = 0x07
REG_FIFO_CONFIG = 0x08

REG_MODE_CONFIG = 0x09
REG_SPO2_CONFIG = 0x0A
REG_LED1_PA = 0x0C
REG_LED2_PA = 0x0D
REG_MULTI_LED_CTRL1 = 0x11
REG_MULTI_LED_CTRL2 = 0x12

REG_TEMP_INTR = 0x1F
REG_TEMP_FRAC = 0x20
REG_TEMP_CONFIG = 0x21
REG_PROX_INT_THRESH = 0x30
REG_REV_ID = 0xFE
REG_PART_ID = 0xFF

# Parameter (Byte-Typ) 
temp_enable = b'\x01'
red_only = b'\x02'
red_pulse_width = b'\x03' # 411 us
led_current_0_mA = b'\x00'
led_current_3_mA = b'\x0F'
led_current_7_mA = b'\x24'
reset = b'\x40'
shutdown = b'\x80'
int_enable_1 = b'\xC0' # A_FULL_EN und PPG_RDY_EN aktiviert
int_enable_2 = b'\x00' # DIE_TEMP_RDY_EN deaktiviert
fifo_config = b'\x4F'  # 0b 0100 1111, d. h. # Sample Avg = 4, fifo rollover = False, FIFO_Almost_FULL = 15
spo2_config = b'\x27'  # # 0b 0 01 001 11, d. h. SPO2_ADC range = 4096nA, SPO2 sample rate = 100Hz, LED pulse-width = 411uS


class MAX30102():
    def __init__(self, address=0x57, scl_Pin_nr=25, sda_Pin_nr=26):
        print('I2C-Adresse:', address)
        self.address = address
        self.i2c = I2C(1, scl=Pin(scl_Pin_nr), sda=Pin(sda_Pin_nr))

        self.reset()
        sleep(1)  # warte 1 s
        reg_data = self.i2c.readfrom_mem(self.address, REG_INTR_STATUS_1, 1)
        print('Interrupt Status 1:', bin(int.from_bytes(reg_data, 'big')))        
        print('Der Wert ist 0b1, wenn zuvor ein Power-Up stattgefunden hat, sonst 0b0.')
        
    def reset(self):
        self.i2c.writeto_mem(self.address, REG_MODE_CONFIG, reset)
        print('MAX30102 resettet')
        
    def shutdown(self):
        self.i2c.writeto_mem(self.address, REG_MODE_CONFIG, shutdown)
        print('MAX30102 resettet')        
        
    def get_temp(self) -> num: # liefert Temperatur des MAX30102
        self.i2c.writeto_mem(self.address, REG_TEMP_CONFIG, temp_enable)        
        # setzt das TEMP_EN (able) - Bit auf 1; leitet damit eine Messung ein;
        # anschließend wird es wieder auf 0 gesetzt
        sleep(0.05) # min 29 ms
        temperature_raw_value = self.i2c.readfrom_mem(self.address, REG_TEMP_INTR, 2)
        # print('Rohwert', temperature_raw_value)
        temperature = temperature_raw_value[0] + temperature_raw_value[1]*0.0625
        # print(temperature, '°C')
        return temperature
    
    def init_red_led(self): # NUR RED-LED
        self.i2c.writeto_mem(self.address, REG_MODE_CONFIG, red_only) # RED ONLY (unbedingt erforderlich, sonst bleiben LEDs aus!!!)
        self.i2c.writeto_mem(self.address, REG_SPO2_CONFIG, red_pulse_width) # Pulsweite 411 us (deutlich heller als bei 69 us)
        
    def set_red_led(self, current): # NUR RED-LED
        self.i2c.writeto_mem(self.address, REG_LED1_PA, current) # Pulsweite 411 us (deutlich heller als bei 69 us)

    def get_reg(self, reg):
        return self.i2c.readfrom_mem(self.address, reg, 1)
    
    def setup(self, led_mode=b'\x03'): # default: RED-LED und IR-LED aktivieren?
        # INTR-Einstellungen:
        # 0xc0 : A_FULL_EN und PPG_RDY_EN = Interrupt werden getriggert wenn
        # FIFO nahezu voll & new fifo data ready
        self.i2c.writeto_mem(self.address, REG_INTR_ENABLE_1, int_enable_1)
        self.i2c.writeto_mem(self.address, REG_INTR_ENABLE_2, int_enable_2)
        
        # FIFO_WR_PTR[4:0] auf 0 setzen
        self.i2c.writeto_mem(self.address, REG_FIFO_WR_PTR, b'\x00')

        # OVF_COUNTER[4:0] auf 0 setzen
        self.i2c.writeto_mem(self.address, REG_OVF_COUNTER, b'\x00')

        # FIFO_RD_PTR[4:0] auf 0 setzen
        self.i2c.writeto_mem(self.address, REG_FIFO_RD_PTR, b'\x00')

        # Sample Avg = 2, fifo rollover = False, FIFO_Almost_FULL = 15
        self.i2c.writeto_mem(self.address, REG_FIFO_CONFIG, fifo_config)

        # 0x02 für HR mode (Red only), 0x03 für SpO2 mode (default), 0x07 multimode LED
        self.i2c.writeto_mem(self.address, REG_MODE_CONFIG, led_mode)

        # 0b 0 01 001 11 für
        # ADC Range Control = 4096 nA, SPO2 sample rate = 100Hz, LED pulse-width = 411uS
        self.i2c.writeto_mem(self.address, REG_SPO2_CONFIG, spo2_config)        

        # wähle Wert von ~7mA für LED1 (Rote LED)
        self.i2c.writeto_mem(self.address, REG_LED1_PA, led_current_7_mA)

        # wähle Wert von ~7mA für LED2 (IR-LED)
        self.i2c.writeto_mem(self.address, REG_LED2_PA, led_current_7_mA)

        print('SETUP fertig mit LED-Mode', led_mode)
        print()
        sleep(1)

    # wenn die Default-Werte geändert werden sollen, benutze...
    def set_config(self, reg, value):
        self.i2c.writeto_mem(self.address, reg, value)
    
    # Wie viele Datensätze (Samples) sind im FIFO-Puffer?
    def get_data_present(self):
        read_ptr = self.i2c.readfrom_mem(self.address, REG_FIFO_RD_PTR, 1)
        read_ptr = int.from_bytes(read_ptr, 'big')
        # print('RD_PTR', read_ptr)
        # read_ptr wird bei jedem Lese-Vorgang inkrementiert,
        # dabei kann ggf. ein Überlauf erfolgen
        
        # to do: manchmal gibt es die Fehlermeldung "[Errno 19] ENODEV"
        write_ptr = self.i2c.readfrom_mem(self.address, REG_FIFO_WR_PTR, 1)
        write_ptr = int.from_bytes(write_ptr, 'big')
        # print('WR_PTR', write_ptr)
        # write_ptr wird bei jedem Schreib-Vorgang inkrementiert,
        # dabei kann ggf. ein Überlauf erfolgen
        if read_ptr == write_ptr:
            return 0
        else:
            num_samples = write_ptr - read_ptr
            # pointer wrap around berücksichtigen:
            if num_samples < 0:
                num_samples += 32
            return num_samples
        
    def read_fifo(self): # ein einziges Sample aus dem FIFO-Puffer lesen
        red_led = None
        ir_led = None

        # jeweils 1 Byte aus den Registern REG_INTR_STATUS_1 und REG_INTR_STATUS_2 lesen,
        # dadurch werden diese Interrupt-Register gelöscht
        reg_INTR1 = self.i2c.readfrom_mem(self.address, REG_INTR_STATUS_1, 1)
        reg_INTR2 = self.i2c.readfrom_mem(self.address, REG_INTR_STATUS_2, 1)

        # 6 Bytes (3 für RED-LED und 3 für IR-LED) aus dem FIFO lesen
        d = list(self.i2c.readfrom_mem(self.address, REG_FIFO_DATA, 6))
        # print('Byte-Liste für 1 Sample', d)
        # aus jeweils 3 Bytes die zugehörige 3-Byte-Zahl formen;
        # nur die unteren 18 Bits benutzen (MSB ist in Bit 17),
        # also mit 0x03FFFF maskieren (vgl. Datasheet S. 15)
        # Die Listenelemente d[0], d[1], ... sind Zahlen!
        red_led = (d[0] << 16 | d[1] << 8 | d[2]) & 0x03FFFF
        ir_led = (d[3] << 16 | d[4] << 8 | d[5]) & 0x03FFFF

        return red_led, ir_led
    
